/**
 *       Title:  PFE bindings for the GNU Multiple
 *               Precision Library, GNU MP 5.0.1
 *        File:  libgmp-pfe.c
 *    Log file:  gmpfr.log
 *  Test files:  libgmp-test.fs, gmpfr-test.fs
 *      Author:  David N. Williams
 *     License:  LGPL
 *     Version:  0.8.0
 *     Started:  January 15, 2011 
 *     Revised:  March 26, 2011
 *
 * Copyright (C) 2011 by David N. Williams
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 *
 * This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU Library General Public License for moref
 * details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA.
 *
 * If you take advantage of the option in the LGPL to put a
 * particular version of this library under the GPL, the author
 * would regard it as polite if you would put any direct
 * modifications under the LGPL as well, and include a copy of
 * this request near the beginning of the modified library
 * source.  A "direct modification" is one that enhances or
 * extends the library in line with its original concept, as
 * opposed to developing a distinct application or library which
 * might use it.
 */


/**********
  Preamble   
 **********/

/**
 * This is an external pfe module, not part of the pfe
 * distribution.
 *
 * This extension word set implements pfe bindings for GMP.
 * Its words are based on the data stack, and have been
 * used to develop a library with separate stacks for
 * multiprecision integers, rationals, and floating-point
 * numbers.
 *
 * The names of Forth binding words are the same as those of the
 * corresponding GMP functions; and the order of Forth data
 * stack inputs is the same as that of the GMP function
 * arguments, read from left to right in the stack effect.  In
 * particular, that means that the addresses of output records
 * are deepest on the input data stack.  All GMP function
 * returns go to the Forth output stack.
 *
 * It is hoped that these conventions, together with the
 * notation described below, make it practical to use the GNU MP
 * manual as documentation for the Forth bindings.  It is well
 * written, and available in both html and pdf format at
 * 
 *   http://gmplib.org/
 */

#define _P4_SOURCE 1
#include <stdio.h>
#include <gmp.h>

#include <pfe/pfe-base.h>
 
#include "libgmp-pfe.h"


/*******************
  Table of contents
 *******************/

/***
 * 0 Notation and terminology
 * 1 General support
 * 2 Integers
 * 2.1  Initialization
 * 2.2  Assignment
 * 2.3  Combined initialization and assignment
 * 2.4  Conversion
 * 2.5  Arithmetic
 * 2.6  Division
 * 2.7  Exponentiation
 * 2.8  Root extraction
 * 2.9  Number theoretics
 * 2.10 Comparison
 * 2.11 Logic and big manipulation
 * 2.12 Input and output
 * 2.13 Random numbers
 * 2.14 Integer export and import
 * 2.15 Miscellaneous
 * 2.16 Special
 *
 * 3 Rationals
 * 3.1 Initialization and assignment
 * 3.2 Conversion
 * 3.3 Arithmetic
 * 3.4 Comparison
 * 3.5 Integer functions
 * 3.6 Input and output
 *
 * 4 Floating point
 * 4.1 Initialization
 * 4.2 Assignment
 * 4.3 Combined initialization and assignment
 * 4.4 Conversion
 * 4.5 Arithmetic
 * 4.6 Comparison
 * 4.7 Input and output
 * 4.8 Miscellaneous
 *
 * 5 Low level
 *
 * 6 Random number state
 * 6.1 Initialization
 * 6.2 Seeding
 * 6.3 Miscellaneous
 *
 * 7 Formatted output
 *
 * 8 Formatted input
 */


/****************************
  0 Notation and terminology
 ****************************/

/**
 * ' prefix:  A common Forth style used to indicate an address.
 *
 * 'b, 'b1, ...:  Address of an initialized, mpz integer
 *   object.  The gmp type is mpz_t.  The corresponding big
 *   numbers are called b, b1, b2, ...  The analogous notation
 *   for quotient (rational number) and floating point objects
 *   is 'q, 'f, etc.  They have gmp types mpq_t and mpf_t.
 *
 * 'r=... or '...:  Address of an mpz result object that is either
 *   already initialized, or about to be initialized by one of
 *   the initialization functions.  The "=..." stands for
 *   assignment of the value of the expression "..." to r.  The
 *   expression may include r itself, as for example with
 *   mpz_addmul().  The '...  notation is short fort 'r=....
 *
 * +p, +den, ...:  Unsigned power, denominator, etc.,
 *   corresponding to the Forth u type (not the Forth +n type).
 */


/*****************
  1 General support
 *****************/

/**
 * gmp_version  ( -- 'zstr )
 */
FCode (p4_gmp_version) { *--SP = (p4ucell) gmp_version; }

/**
 * mp-bits_per_limb  ( -- mp_bits_per_limb )
 */
FCode (p4_mp_bits_per_limb) { *--SP = mp_bits_per_limb; }


/************
  2 Integers
 ************/

/**
 * /MPZ  ( -- sizeof[mpz_t] )
 */

/********************
  2.1 Initialization
 ********************/

/**
 * mpz_init  ( 'b=0 -- )
 */
FCode (p4_mpz_init)
{
  mpz_init ( (mpz_ptr) *SP++ );
}

/**
 * mpz_clear  ( 'b -- )
 */
FCode (p4_mpz_clear)
{
  mpz_clear ( (mpz_ptr) *SP++ );
}

/****************
  2.2 Assignment
 ****************/

/**
 * mpz_set  ( 'b=b1 'b1 -- )
 */
FCode (p4_mpz_set)
{
  mpz_set ( (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 2;
}

/**
 * mpz_set_ui  ( 'b=u u -- )
 */
FCode (p4_mpz_set_ui)
{
  mpz_set_ui ( (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 2;
}

/**
 * mpz_set_si  ( 'b=n n -- )
 */
FCode (p4_mpz_set_si)
{
  mpz_set_si ( (mpz_ptr) SP[1], (slong) SP[0] );
  SP += 2;
}

/**
 * mpz_set_d  ( 'b=x f: x -- )
 */
FCode (p4_mpz_set_d)
{
  mpz_set_d ( (mpz_ptr) *SP++, *FP++ );
}

/**
 * mpz_set_q  ( 'b=q 'q -- )
 */
FCode (p4_mpz_set_q)
{
  mpz_set_q ( (mpz_ptr) SP[1], (mpq_ptr) SP[0] );
  SP += 2;
}

/**
 * mpz_set_f  ( 'b=f 'f -- )
 */
FCode (p4_mpz_set_f)
{
  mpz_set_f ( (mpz_ptr) SP[1], (mpf_ptr) SP[0] );
  SP += 2;
}

/**
 * mpz_set_str  ( 'b=conversion 'bzstr base|0 -- flag )
 */ 
FCode (p4_mpz_set_str)
{
  int nflag = mpz_set_str ( (mpz_ptr) SP[2], (char*) SP[1], (int) SP[0]);
  SP += 2;
  *SP = nflag;
}

/**
 * mpz_swap  ( 'b1 'b2 -- )
 */
FCode (p4_mpz_swap)
{
  mpz_swap ( (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 2;
}

/********************************************
  2.3 Combined initialization and assignment
 ********************************************/

/**
 * mpz_init_set  ( 'r=b 'b -- )
 */
FCode (p4_mpz_init_set)
{
  mpz_init_set ( (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 2;
}

/**
 * mpz_init_set_ui  ( 'b=u u -- )
 *
 * Not to be used when 'r is already initialized!
 */
FCode (p4_mpz_init_set_ui)
{
  mpz_init_set_ui ( (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 2;
}

/**
 * mpz_init_set_si  ( 'b=n n -- )
 */
FCode (p4_mpz_init_set_si)
{
  mpz_init_set_si ( (mpz_ptr) SP[1], (slong) SP[0] );
  SP += 2;
}

/**
 * mpz_init_set_d  ( 'b=x f: x -- )
 */
FCode (p4_mpz_init_set_d)
{
  mpz_init_set_d ( (mpz_ptr) *SP++, *FP++ );
}

/**
 * mpz_init_set_str  ( 'b=conversion 'bzstr base -- flag  ) 
 */ 
FCode (p4_mpz_init_set_str)
{
  int nflag = mpz_init_set_str ( (mpz_ptr) SP[2], (char*) SP[1],
    (int) SP[0]);
  SP += 2;
  *SP = nflag;
}

/****************
  2.4 Conversion
 ****************/

/**
 * mpz_get_ui  ( 'b -- |b| ) \ truncated
*/
FCode (p4_mpz_get_ui)
{
  *SP = mpz_get_ui ((mpz_ptr) *SP);
}

/**
 * mpz_get_si  ( 'b -- b ) \ truncated
*/
FCode (p4_mpz_get_si)
{
  *SP = mpz_get_si ((mpz_ptr) *SP);
}

/**
 * mpz_get_d  ( 'b -- f: conversion )
*/
FCode (p4_mpz_get_d)
{
  *--FP = mpz_get_d ((mpz_ptr) *SP++);
}

/**
 * mpz_get_d_2exp  ( 'exp 'b -- f: frac )
*/
FCode (p4_mpz_get_d_2exp)
{
  *--FP = mpz_get_d_2exp ((signed long int*) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpz_get_str  ( 'buf|0 base 'b -- 'zstr )
*/
FCode (p4_mpz_get_str)
{
//  char *p = mpz_get_str ( (char*) SP[2], (int) SP[1], (mpz_ptr) SP[0]);
  char *p = mpz_get_str ( (void*) SP[2], (int) SP[1], (void*) SP[0]);
  SP += 2;
  *SP = (p4ucell) p;
}

/****************
  2.5 Arithmetic
 ****************/

/**
 * mpz_add  ( 'b1+b2 'b1 'b2 -- )
 */
FCode (p4_mpz_add)
{
  mpz_add ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_add_ui  ( 'b+u 'b u -- )
 */
FCode (p4_mpz_add_ui)
{
  mpz_add_ui ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 3;
}

/**
 * mpz_sub  ( 'b1-b2 'b1 'b2 -- )
 */
FCode (p4_mpz_sub)
{
  mpz_sub ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_sub_ui  ( 'b-u 'b u -- )
 */
FCode (p4_mpz_sub_ui)
{
  mpz_sub_ui ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 3;
}

/**
 * mpz_ui-sub  ( 'u-b u 'b -- )
 */
FCode (p4_mpz_ui_sub)
{
  mpz_ui_sub ( (mpz_ptr) SP[2], (ulong) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_mul  ( 'b1*b2 'b1 'b2 -- )
 */
FCode (p4_mpz_mul)
{
  mpz_mul ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_mul_si  ( 'b*n 'b n -- )
 */
FCode (p4_mpz_mul_si)
{
  mpz_mul_si ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (slong) SP[0] );
  SP += 3;
}

/**
 * mpz_mul_ui  ( 'b*u 'b u -- )
 */
FCode (p4_mpz_mul_ui)
{
  mpz_mul_ui ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 3;
}

/**
 * mpz_addmul  ( 'b=b+b1*b2 'b1 'b2 -- )
 */
FCode (p4_mpz_addmul)
{
  mpz_addmul ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_addmul_ui  ( 'b=b+b1*u 'b1 u -- )
 */
FCode (p4_mpz_addmul_ui)
{
  mpz_addmul_ui ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 3;
}

/**
 * mpz_submul  ( 'b=b-b1*b2 'b1 'b2 -- )
 */
FCode (p4_mpz_submul)
{
  mpz_submul ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 3;
}

/**
 * mpz_submul_ui  ( 'b=b-b*u 'b u -- )
 */
FCode (p4_mpz_submul_ui)
{
  mpz_submul_ui ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0] );
  SP += 3;
}

/**
 * mpz_mul_2exp  ( 'b*2^p 'b +p -- )
 */
FCode (p4_mpz_mul_2exp)
{
  mpz_mul_2exp ( (mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0] );
  SP += 3;
}

/**
 * mpz_neg  ( '-b 'b -- )
 */
FCode (p4_mpz_neg)
{
  mpz_neg ( (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 2;
}

/**
 * mpz_abs  ( '|b| 'b -- )
 */
FCode (p4_mpz_abs)
{
  mpz_abs ( (mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  SP += 2;
}

/**************
  2.6 Division
 **************/

/**
 * mpz_cdiv_q  ( 'quo 'num 'den -- )
 */
FCode (p4_mpz_cdiv_q)
{
  mpz_cdiv_q ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_cdiv_r  ( 'rem 'num 'den -- )
 */
FCode (p4_mpz_cdiv_r)
{
  mpz_cdiv_r ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_cdiv_qr  ( 'quo 'rem 'num 'den -- )
 */
FCode (p4_mpz_cdiv_qr)
{
  mpz_cdiv_qr ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_cdiv_q_ui  ( 'quo 'num +den -- |rem| )
 */
FCode (p4_mpz_cdiv_q_ui)
{
  ulong absr =
    mpz_cdiv_q_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_cdiv-R_ui  ( 'rem 'num +den -- |rem| )
 */
FCode (p4_mpz_cdiv_r_ui)
{
  ulong absr =
    mpz_cdiv_r_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_cdiv_qr_ui  ( 'quo 'rem 'num +den -- |rem| )
 */
FCode (p4_mpz_cdiv_qr_ui)
{
  ulong absr =
    mpz_cdiv_qr_ui ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
      (ulong) SP[0]);
  SP += 3;
  *SP = absr;
}

/**
 * mpz_cdiv_ui  ( 'num +den -- |rem| )
 */
FCode (p4_mpz_cdiv_ui)
{
  ulong absr = mpz_cdiv_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = absr;
}

/**
 * mpz_cdiv_q_2exp  ( 'quo 'num +p -- )
 */
FCode (p4_mpz_cdiv_q_2exp)
{
  mpz_cdiv_q_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_cdiv_r_2exp  ( 'rem 'num +p -- )
 */
FCode (p4_mpz_cdiv_r_2exp)
{
  mpz_cdiv_r_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_fdiv_q  ( 'quo 'num 'den -- )
 */
FCode (p4_mpz_fdiv_q)
{
  mpz_fdiv_q ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_fdiv_r  ( 'rem 'num 'den -- )
 */
FCode (p4_mpz_fdiv_r)
{
  mpz_fdiv_r ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_fdiv_qr  ( 'quo 'rem 'num 'den -- )
 */
FCode (p4_mpz_fdiv_qr)
{
  mpz_fdiv_qr ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_fdiv_q_ui  ( 'quo 'num +den -- |rem| )
 */
FCode (p4_mpz_fdiv_q_ui)
{
  ulong absr =
    mpz_fdiv_q_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_fdiv_r_ui  ( 'reem 'num +den -- |rem| )
 */
FCode (p4_mpz_fdiv_r_ui)
{
  ulong absr =
    mpz_fdiv_r_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_fdiv_qr_ui  ( 'quo 'rem 'num +den -- |rem| )
 */
FCode (p4_mpz_fdiv_qr_ui)
{
  ulong absr =
    mpz_fdiv_qr_ui ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
      (ulong) SP[0]);
  SP += 3;
  *SP = absr;
}

/**
 * mpz_fdiv_ui  ( 'num +den -- |rem| )
 */
FCode (p4_mpz_fdiv_ui)
{
  ulong absr = mpz_fdiv_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = absr;
}

/**
 * mpz_fdiv_q_2exp  ( 'quo 'num +den -- )
 */
FCode (p4_mpz_fdiv_q_2exp)
{
  mpz_fdiv_q_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_fdiv_r_2exp  ( 'rem 'num +den -- )
 */
FCode (p4_mpz_fdiv_r_2exp)
{
  mpz_fdiv_r_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_tdiv_q  ( 'quo 'num 'den -- )
 */
FCode (p4_mpz_tdiv_q)
{
  mpz_tdiv_q ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_tdiv_r  ( 'rem 'num 'den -- )
 */
FCode (p4_mpz_tdiv_r)
{
  mpz_tdiv_r ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_tdiv_qr  ( 'quo 'rem 'num 'den -- )
 */
FCode (p4_mpz_tdiv_qr)
{
  mpz_tdiv_qr ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_tdiv_q_ui  ( 'quo 'num +den -- |rem| )
 */
FCode (p4_mpz_tdiv_q_ui)
{
  ulong absr =
    mpz_tdiv_q_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_tdiv_r_ui  ( 'rem 'num +den -- |rem| )
 */
FCode (p4_mpz_tdiv_r_ui)
{
  ulong absr =
    mpz_tdiv_r_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = absr;
}

/**
 * mpz_tdiv_qr_ui  ( 'quo 'rem 'num +den -- |rem| )
 */
FCode (p4_mpz_tdiv_qr_ui)
{
  ulong absr = mpz_tdiv_qr_ui ((mpz_ptr) SP[3], (mpz_ptr) SP[2],
      (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
  *SP = absr;
}

/**
 * mpz_tdiv_ui  ( 'num +den -- |rem| )
 */
FCode (p4_mpz_tdiv_ui)
{
  ulong absr = mpz_tdiv_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = absr;
}

/**
 * mpz_tdiv_q_2exp  ( 'quo 'num +den -- )
 */
FCode (p4_mpz_tdiv_q_2exp)
{
  mpz_tdiv_q_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_tdiv_r_2exp  ( 'rem 'num +den -- )
 */
FCode (p4_mpz_tdiv_r_2exp)
{
  mpz_tdiv_r_2exp ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpz_mod  ( 'rem 'num 'den -- )
 */
FCode (p4_mpz_mod)
{
  mpz_mod ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_mod_ui  ( 'rem 'num +den -- +mod )
 */
FCode (p4_mpz_mod_ui)
{
  ulong mod =
    mpz_mod_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = mod;
}

/**
 * mpz_divexact  ( 'quo 'num 'den -- )
 */
FCode (p4_mpz_divexact)
{
  mpz_divexact ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_divexact_ui  ( 'quo 'num +den -- )
 */
FCode (p4_mpz_divexact_ui)
{
  mpz_divexact_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpz_divisible_p  ( 'num 'den -- flag )
 */
FCode (p4_mpz_divisible_p)
{
  int flag = mpz_divisible_p ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  *++SP = flag ? ~0 : 0 ;
}

/**
 * mpz_divisible_ui_p  ( 'num +den -- flag )
 */
FCode (p4_mpz_divisible_ui_p)
{
  int flag = mpz_divisible_ui_p ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = flag ? ~0 : 0 ;
}

/**
 * mpz_divisible_2exp_p  ( 'num +p -- flag )
 */
FCode (p4_mpz_divisible_2exp_p)
{
  int flag = mpz_divisible_2exp_p ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  *++SP = flag ? ~0 : 0 ;
}

/**
 * mpz_congruent_p  ( 'num 'c 'den -- flag )
 */
FCode (p4_mpz_congruent_p)
{
  int flag = mpz_congruent_p ((mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_congruent_ui_p  ( 'num +c +den -- flag )
 */
FCode (p4_mpz_congruent_ui_p)
{
  int flag = mpz_congruent_ui_p ((mpz_ptr) SP[2], (ulong) SP[1],
    (ulong) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_congruent_2exp_p  ( 'num 'c +p -- flag )
 */
FCode (p4_mpz_congruent_2exp_p)
{
  int flag = mpz_congruent_2exp_p ((mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mp_bitcnt_t) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0 ;
}

/********************
  2.7 Exponentiation
 ********************/

/**
 * mpz_powm  ( 'b 'base 'p 'mod -- )
 */
FCode (p4_mpz_powm)
{
  mpz_powm ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_powm_ui  ( 'b 'base +p 'mod -- )
 */
FCode (p4_mpz_powm_ui)
{
  mpz_powm_ui ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (ulong) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_powm_sec  ( 'b 'base 'p 'mod -- )
 */
FCode (p4_mpz_powm_sec)
{
  mpz_powm_sec ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 4;
}

/**
 * mpz_pow_ui  ( 'b 'base +p -- )
 */
FCode (p4_mpz_pow_ui)
{
  mpz_pow_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpz_ui_pow_ui  ( 'b base +p -- )
 */
FCode (p4_mpz_ui_pow_ui)
{
  mpz_ui_pow_ui ((mpz_ptr) SP[2], (ulong) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**********************
  2.8  Root extraction
 **********************/

/**.
 * mpz_root  ( 'root 'b n -- flag )
 */
FCode (p4_mpz_root)
{
  int flag = mpz_root ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_rootrem  ( 'root 'rem 'b u -- )
 */
FCode (p4_mpz_rootrem)
{
  mpz_rootrem ((mpz_ptr) SP[3], (mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (ulong) SP[0]);
  SP += 4;
}

/**
 * mpz_sqrt  ( 'root 'b -- )
 */
FCode (p4_mpz_sqrt)
{
  mpz_sqrt ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpz_sqrtrem  ( 'root 'rem 'b -- )
 */
FCode (p4_mpz_sqrtrem)
{
  mpz_sqrtrem ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_perfect_power_p  ( 'b -- flag )
 */
FCode (p4_mpz_perfect_power_p)
{
  int flag = mpz_perfect_power_p ((mpz_ptr) SP[0]);
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_perfect_square  ( 'b -- flag )
 */
FCode (p4_mpz_perfect_square_p)
{
  int flag = mpz_perfect_square_p ((mpz_ptr) SP[0]);
  *SP = flag ? ~0 : 0 ;
}

/************************
  2.9  Number theoretics
 ************************/

/**
 * mpz_probab_prime_p  ( 'b #reps -- 2|1|0 )
 */
FCode (p4_mpz_probab_prime_p)
{
  int probcase  = mpz_probab_prime_p ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = probcase ;
}

/**
 * mpz_nextprime  ( 'nextprime 'prime -- )
 */
FCode (p4_mpz_nextprime)
{
  mpz_nextprime ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpz_gcd  ( '+gcd 'b1 'b2 -- )
 */
FCode (p4_mpz_gcd)
{
  mpz_gcd ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_gcd_ui  ( '+gcd|0 'b1 u2 -- +gcd )
 */
FCode (p4_mpz_gcd_ui)
{
  ulong gcd = mpz_gcd_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = gcd;
}

/**
 * mpz_gcdext  ( 'g 's 't|0 'a 'b -- )
 */
FCode (p4_mpz_gcdext)
{
  mpz_gcdext ((mpz_ptr) SP[4], (mpz_ptr) SP[3], (mpz_ptr) SP[2],
    (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 5;
}

/**
 * mpz_lcm  ( '+lcm 'b1 'b2 -- )
 */
FCode (p4_mpz_lcm)
{
  mpz_lcm ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_lcm_ui  ( '+lcm 'b1 u2 -- )
 */
FCode (p4_mpz_lcm_ui)
{
  mpz_lcm_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpz_invert  ( 'inv 'b1 'b2 -- flag )
 */
FCode (p4_mpz_invert)
{
  int flag = mpz_invert ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_jacobi  ( 'a 'b -- jacsym )
 */
FCode (p4_mpz_jacobi)
{
  int jacsym = mpz_jacobi ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  *++SP = jacsym ;
}

/**
 * mpz_legendre  ( 'a 'prime -- legsym )
 */
FCode (p4_mpz_legendre)
{
  int legsym = mpz_legendre ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  *++SP = legsym ;
}

/**
 * mpz_kronecker  ( 'a 'b -- jacsymext )
 */
FCode (p4_mpz_kronecker)
{
  int jacsymext = mpz_kronecker ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  *++SP = jacsymext ;
}

/**
 * mpz_kronecker_si  ( 'a n -- jacsymext )
 */
FCode (p4_mpz_kronecker_si)
{
  int jacsymext = mpz_kronecker_si ((mpz_ptr) SP[1], (slong) SP[0]);
  *++SP = jacsymext ;
}

/**
 * mpz_kronecker_ui  ( 'a u -- jacsymext )
 */
FCode (p4_mpz_kronecker_ui)
{
  int jacsymext = mpz_kronecker_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = jacsymext ;
}

/**
 * mpz_si_kronecker  ( n 'b -- jacsymext )
 */
FCode (p4_mpz_si_kronecker)
{
  int jacsymext = mpz_si_kronecker ((slong) SP[1], (mpz_ptr) SP[0]);
  *++SP = jacsymext ;
}

/**
 * mpz_ui_kronecker  ( u 'b -- jacsymext )
 */
FCode (p4_mpz_ui_kronecker)
{
  int jacsymext = mpz_ui_kronecker ((slong) SP[1], (mpz_ptr) SP[0]);
  *++SP = jacsymext ;
}

/**
 * mpz_remove  ( 'bred 'b 'f -- #removed )
 */
FCode (p4_mpz_remove)
{
  mp_bitcnt_t nremoved = mpz_remove ((mpz_ptr) SP[2], (mpz_ptr) SP[1],
    (mpz_ptr) SP[0]);
  SP += 2 ;
  *SP = (p4cell) nremoved;
}

/**
 * mpz_fac_ui  ( 'u! u -- )
 */
FCode (p4_mpz_fac_ui)
{
  mpz_fac_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2 ;
}

/**
 * mpz_bin_ui  ( 'bin 'n +k -- )
 */
FCode (p4_mpz_bin_ui)
{
  mpz_bin_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3 ;
}

/**
 * mpz_bin_uiui  ( 'bin +n +k -- )
 */
FCode (p4_mpz_bin_uiui)
{
  mpz_bin_uiui ((mpz_ptr) SP[2], (ulong) SP[1], (ulong) SP[0]);
  SP += 3 ;
}

/**
 * mpz_fib_ui  ( 'fn +n -- )
 */
FCode (p4_mpz_fib_ui)
{
  mpz_fib_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2 ;
}

/**
 * mpz_fib2_ui  ( 'fn 'fnsub1 +n -- )
 */
FCode (p4_mpz_fib2_ui)
{
  mpz_fib2_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3 ;
}

/**
 * mpz_lucnum_ui  ( 'ln +n -- )
 */
FCode (p4_mpz_lucnum_ui)
{
  mpz_lucnum_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 2 ;
}

/**
 * mpz_lucnum2_ui  ( 'ln 'lnsub1 +n -- )
 */
FCode (p4_mpz_lucnum2_ui)
{
  mpz_lucnum2_ui ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (ulong) SP[0]);
  SP += 3 ;
}

/*****************
  2.10 Comparison
 *****************/

/**
 * mpz_cmp  ( 'b1 'b2 -- +n|0|-n )
 * +n: b1 > b2
 *  0: b1 = b2
 * -n: b1 < b2
 */
FCode (p4_mpz_cmp)
{
  int n = mpz_cmp ((mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  *++SP = n;
}

/**
 * mpz_cmp_d  ( 'b f: x -- s: +n|0|-n )
 * +n: b > x
 *  0: b = x
 * -n: b < x
 */
FCode (p4_mpz_cmp_d)
{
  *SP = mpz_cmp_d ((mpz_ptr) *SP, *FP++ );
}

/**
 * mpz_cmp_si  ( 'b n -- +n'|0|-n' )
 * +n: b > n
 *  0: b = n
 * -n: b < n
 */
FCode (p4_mpz_cmp_si)
{
  int n = mpz_cmp_si ((mpz_ptr) SP[1], (slong) SP[0]);
  *++SP = n;
}

/**
 * mpz_cmp_ui  ( 'b u -- +n'|0|-n' )
 * +n: b > u
 *  0: b = u
 * -n: b < u
 */
FCode (p4_mpz_cmp_ui)
{
  int n = mpz_cmp_ui ((mpz_ptr) SP[1], (ulong) SP[0]);
  *++SP = n;
}

/**
 * mpz_cmpabs  ( 'b1 'b2 -- +n|0|-n )
 * +n: |b1| > |b2|
 *  0: |b1| = |b2|
 * -n: |b1| < |b2|
 */
FCode (p4_mpz_cmpabs)
{
  int n = mpz_cmpabs ((mpz_ptr) SP[1], (mpz_ptr) SP[0] );
  *++SP = n;
}

/**
 * mpz_cmpabs_d  ( 'b f: x -- s: +n|0|-n )
 * +n: |b| > |x|
 *  0: |b| = |x|
 * -n: |b| < |x|
 */
FCode (p4_mpz_cmpabs_d)
{
  *SP = mpz_cmpabs_d ((mpz_ptr) *SP, *FP++ );
}

/**
 * mpz_cmp-abs_ui  ( 'b u -- +n|0|-n )
 * +n: |b| > u
 *  0: |b| = u
 * -n: |b| < u
 */
FCode (p4_mpz_cmpabs_ui)
{
  int n = mpz_cmpabs_ui ((mpz_ptr) SP[1], (ulong) SP[0] );
  *++SP = n;
}

/**
 * mpz_sgn  ( 'b -- s: +1|0|-1 )
 */
FCode (p4_mpz_sgn)
{
  *SP = mpz_sgn ((mpz_ptr) *SP);
}

/*********************************
  2.11 Logic and bit manipulation
 *********************************/

/**
 * mpz_and  ( 'b1.and.b2 'b1 'b2 -- )
 */
FCode (p4_mpz_and)
{
  mpz_and ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_ior  ( 'b1.or.b2 'b1 'b2 -- )
 */
FCode (p4_mpz_ior)
{
  mpz_ior ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_xor  ( 'b1.xor.b2 'b1 'b2 -- )
 */
FCode (p4_mpz_xor)
{
  mpz_xor ((mpz_ptr) SP[2], (mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 3;
}

/**
 * mpz_comp  ( '~b 'b -- )
 */
FCode (p4_mpz_com)
{
  mpz_com ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpz_popcount  ( 'b -- #1bits )
 */
FCode (p4_mpz_popcount)
{
  *SP = (p4ucell) mpz_popcount ((mpz_ptr) *SP);
}

/**
 * mpz_hamdist  ( 'b1 'b2 -- hamdist )
 */
FCode (p4_mpz_hamdist)
{
  ulong ham = mpz_hamdist ((mpz_ptr) SP[1], (mpz_ptr) SP[0]);
  *++SP = ham;
}

/**
 * mpz_scan0  ( 'b startbit# -- first0bit  )
 */
FCode (p4_mpz_scan0)
{
  ulong first0bit = mpz_scan0 ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  *++SP = first0bit;
}

/**
 * mpz_scan1  ( 'b startbit# -- first1bit  )
 */
FCode (p4_mpz_scan1)
{
  ulong first1bit = mpz_scan1 ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  *++SP = first1bit;
}

/**
 * mpz_setbit  ( 'b bit# -- )
 */
FCode (p4_mpz_setbit)
{
  mpz_setbit ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/**
 * mpz_clrbit  ( 'b bit# -- )
 */
FCode (p4_mpz_clrbit)
{
  mpz_clrbit ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/**
 * mpz_combit  ( 'b bit# -- )
 */
FCode (p4_mpz_combit)
{
  mpz_combit ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/**
 * mpz_tstbit  ( 'b bit# -- flag  )
 */
FCode (p4_mpz_tstbit)
{
  ulong flag = mpz_tstbit ((mpz_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  *++SP = flag ? ~0 : 0 ;
}

/***********************
  2.12 Input and output
 ***********************/

/**
 * mpz_out_str  ( fid base 'b -- #bytes|0  )
 */
FCode (p4_mpz_out_str)
{
  size_t numbytes = mpz_out_str ( ((p4_File*)SP[2])->f, (int) SP[1],
    (mpz_ptr) SP[0] );
  SP += 2;
  *SP = numbytes ;
}

/**
 * mpz_inp_str  ( 'b fid base|0 -- #bytes|0  )
 */
FCode (p4_mpz_inp_str)
{
  size_t numbytes = mpz_inp_str ((mpz_ptr) SP[2], ((p4_File*)SP[1])->f,
    (int) SP[0]);
  SP += 2;
  *SP = numbytes ;
}

/**
 * mpz_out_raw  ( fid 'b -- #bytes|0  )
 */
FCode (p4_mpz_out_raw)
{
  size_t numbytes = mpz_out_raw ( ((p4_File*)SP[1])->f, (mpz_ptr) SP[0] );
  SP += 1;
  *SP = numbytes ;
}

/**
 * mpz_inp_raw  ( 'b fid -- #bytes|0  )
 */
FCode (p4_mpz_inp_raw)
{
  size_t numbytes = mpz_inp_raw ( (mpz_ptr) SP[1], ((p4_File*)SP[0])->f );
  SP += 1;
  *SP = numbytes ;
}

/*********************
  2.13 Random numbers
 *********************/

/**
 * mpz_urandomb  ( 'b=ran 'state +n -- )
 */
FCode (p4_mpz_urandomb)
{
  mpz_urandomb ((mpz_ptr) SP[2], (void*) SP[1], (mp_bitcnt_t) SP[0]);
 SP += 3;
}

/**
 * mpz_urandomm  ( 'b=ran 'state '+n -- )
 */
FCode (p4_mpz_urandomm)
{
  mpz_urandomm ((mpz_ptr) SP[2], (void*) SP[1], (mpz_ptr) SP[0]);
 SP += 3;
}

/**
 * mpz_rrandomb  ( 'b=ran 'state +n -- )
 */
FCode (p4_mpz_rrandomb)
{
  mpz_rrandomb ((mpz_ptr) SP[2], (void*) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/********************************
  2.14 Integer export and import
 ********************************/

/**
 * mpz_import  ( 'b count order size endian nails 'source -- )
 */
FCode (p4_mpz_import)
{
  mpz_import ((mpz_ptr) SP[6], (size_t) SP[5], (int) SP[4],
    (size_t) SP[3], (int) SP[2], (size_t) SP[1], (const void*) SP[0]);
  SP += 7;
}

/**
 * mpz_export  ( 'target|0 'count order size endian nails 'b -- 'target' )
 */
FCode (p4_mpz_export)
{
void* targp = mpz_export ((void*) SP[6], (size_t*) SP[5], (int) SP[4],
    (size_t) SP[3], (int) SP[2], (size_t) SP[1], (mpz_ptr) SP[0]);
  SP += 6;
  *SP = (p4ucell) targp;
}

/********************
  2.15 Miscellaneous
 ********************/

/**
 * mpz_fits_ulong_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_ulong_p)
{
  ulong flag =  mpz_fits_ulong_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_fits_slong_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_slong_p)
{
  ulong flag =  mpz_fits_slong_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_fits_uint_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_uint_p)
{
  ulong flag =  mpz_fits_uint_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_fits_sint_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_sint_p)
{
  ulong flag =  mpz_fits_sint_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_fits_ushort_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_ushort_p)
{
  ulong flag =  mpz_fits_ushort_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_fits_sshort_p  ( 'b -- flag )
 */
FCode (p4_mpz_fits_sshort_p)
{
  ulong flag =  mpz_fits_sshort_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0 ;
}

/**
 * mpz_odd_p  ( 'b -- flag )
 */
FCode (p4_mpz_odd_p)
{
  ulong flag =  mpz_odd_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_even_p  ( 'b -- flag )
 */
FCode (p4_mpz_even_p)
{
  ulong flag =  mpz_even_p ((mpz_ptr) *SP);
  *SP = flag ? ~0 : 0;
}

/**
 * mpz_sizeinbase  ( 'b base -- #digits )
 */
FCode (p4_mpz_sizeinbase)
{
  size_t size =  mpz_sizeinbase ((mpz_ptr) SP[1], (int) SP[0]);
  *++SP = size;
}

/**************
  2.16 Special
 **************/

/**
 * mpz_getlimbn  ( 'b +n -- limb )
 */
FCode (p4_mpz_getlimbn)
{
  mp_limb_t limb = mpz_getlimbn ((mpz_ptr) SP[1], (mp_size_t) SP[0]);
  *++SP = limb;
}

/**
 * mpz_size  ( 'b -- #limbs )
 */
FCode (p4_mpz_size)
{
  *SP = mpz_size ((mpz_ptr) *SP);
}


/*************
  3 Rationals
 *************/

/**
 * /MPQ  ( -- sizeof[mpq_t] )
 */

/**
 * mpq_canonicalize  ( 'q -- )
 */
FCode (p4_mpq_canonicalize)
{
  mpq_canonicalize ((mpq_ptr) *SP++);
}

/***********************************
  3.1 Initialization and assignment
 ***********************************/

/**
 * mpq_init  ( 'q -- )
 */
FCode (p4_mpq_init)
{
  mpq_init ((mpq_ptr) *SP++);
}

/**
 * mpq_clear  ( 'q -- )
 */
FCode (p4_mpq_clear)
{
  mpq_clear ((mpq_ptr) *SP++);
}

/**
 * mpq_set  ( 'q=q1 'q1 -- )
 */
FCode (p4_mpq_set)
{
  mpq_set ((mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 2;
}

/**
 * mpq_set_z  ( 'q=b 'b -- )
 */
FCode (p4_mpq_set_z)
{
  mpq_set_z ((mpq_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpq_set_ui  ( 'q=u +num +den -- )
 */
FCode (p4_mpq_set_ui)
{
  mpq_set_ui ((mpq_ptr) SP[2], (ulong) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpq_set_si  ( 'q=num/den num +den -- )
 */
FCode (p4_mpq_set_si)
{
  mpq_set_si ((mpq_ptr) SP[2], (slong) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpq_set_str  ( 'q 'qzstr base|0 -- flag )
 */ 
FCode (p4_mpq_set_str)
{
  int nflag = mpq_set_str ((mpq_ptr) SP[2], (char*) SP[1], (int) SP[0]);
  SP += 2;
  *SP = nflag;
}

/**
 * mpq_swap  ( 'q1 'q2 -- )
 */
FCode (p4_mpq_swap)
{
  mpq_swap ((mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 2;
}


/****************
  3.2 Conversion
 ****************/

/**
 * mpq_get_d  ( 'q -- f: q ) 
 */ 
FCode (p4_mpq_get_d)
{
  *--FP = mpq_get_d ((mpq_ptr) *SP++);
}

/**
 * mpq_set_d  ( 'q=x f: x -- ) 
 */ 
FCode (p4_mpq_set_d)
{
  mpq_set_d ((mpq_ptr) *SP++, *FP++);
}

/**
 * mpq_set_f  ( 'q=f 'f -- ) 
 */ 
FCode (p4_mpq_set_f)
{
  mpq_set_f ((mpq_ptr) SP[1], (mpf_ptr)SP[0]);
  SP += 2;
}

/**
 * mpq_get_str  ( 'qzstr|0 base 'q -- 'qzstr|'qzstr' ) 
 */ 
FCode (p4_mpq_get_str)
{
  char * convstr = mpq_get_str ((char *) SP[2], (int) SP[1],
    (mpq_ptr) SP[0]);
  SP += 2;
  *SP = (p4ucell) convstr;
}

/****************
  3.3 Arithmetic
 ****************/

/**
 * mpq_add  ( 'q1+q2 'q1 'q2 -- )
 */
FCode (p4_mpq_add)
{
  mpq_add ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 3;
}

/**
 * mpq_sub  ( 'q1-q2 'q1 'q2 -- )
 */
FCode (p4_mpq_sub)
{
  mpq_sub ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 3;
}

/**
 * mpq_mul  ( 'q1*q2 'q1 'q2 -- )
 */
FCode (p4_mpq_mul)
{
  mpq_mul ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 3;
}

/**
 * mpq_mul_2exp  ( 'q1*2^p 'q1 +p -- )
 */
FCode (p4_mpq_mul_2exp)
{
  mpq_mul_2exp ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpq_div  ( 'q1/q2 'q1 'q2 -- )
 */
FCode (p4_mpq_div)
{
  mpq_div ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 3;
}

/**
 * mpq_div_2exp  ( 'q1/2^p 'q1 +p -- )
 */
FCode (p4_mpq_div_2exp)
{
  mpq_div_2exp ((mpq_ptr) SP[2], (mpq_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpq_neg  ( '-q 'q -- )
 */
FCode (p4_mpq_neg)
{
  mpq_neg ((mpq_ptr) SP[1], (mpq_ptr) SP[0] );
  SP += 2;
}

/**
 * mpq_abs  ( '|q| 'q -- )
 */
FCode (p4_mpq_abs)
{
  mpq_abs ((mpq_ptr) SP[1], (mpq_ptr) SP[0] );
  SP += 2;
}

/**
 * mpq_inv  ( '1/q 'q -- )
 */
FCode (p4_mpq_inv)
{
  mpq_inv ((mpq_ptr) SP[1], (mpq_ptr) SP[0] );
  SP += 2;
}

/****************
  3.4 Comparison
 ****************/

/**
 * mpq_cmp  ( 'q1 'q2 -- cmpcase ) 
 */ 
FCode (p4_mpq_cmp)
{
  int cmpcase = mpq_cmp ((mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  *++SP = cmpcase;
}

/**
 * mpq_cmp_ui  ( 'q1 +num2 +den2 -- cmpcase ) 
 */ 
FCode (p4_mpq_cmp_ui)
{
  int cmpcase = mpq_cmp_ui ((mpq_ptr) SP[2], (ulong) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = cmpcase;
}

/**
 * mpq_cmp_si  ( 'q1 num2 +den2 -- cmpcase ) 
 */ 
FCode (p4_mpq_cmp_si)
{
  int cmpcase = mpq_cmp_si ((mpq_ptr) SP[2], (slong) SP[1], (ulong) SP[0]);
  SP += 2;
  *SP = cmpcase;
}

/**
 * mpq_sgn  ( 'q -- 1|0|-1 ) 
 */ 
FCode (p4_mpq_sgn)
{
  *SP = mpq_sgn ((mpq_ptr) *SP);
}

/**
 * mpq_equal  ( 'q1 'q2 -- flag ) 
 */ 
FCode (p4_mpq_equal)
{
  int flag = mpq_equal ((mpq_ptr) SP[1], (mpq_ptr) SP[0]);
  *SP++ = flag ? ~0 : 0;
}

/***********************
  3.5 Integer functions
 ***********************/

/**
 * mpq_numref  ( 'q -- 'num ) 
 */ 
FCode (p4_mpq_numref)
{
  *SP = (p4ucell) mpq_numref ((mpq_ptr) *SP);
}

/**
 * mpq_denref  ( 'q -- 'den ) 
 */ 
FCode (p4_mpq_denref)
{
  *SP = (p4ucell) mpq_denref ((mpq_ptr) *SP);
}

/**
 * mpq_get_num  ( 'num 'q --  ) 
 */ 
FCode (p4_mpq_get_num)
{
  mpq_get_num ((mpz_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 2;
}

/**
 * mpq_get_den  ( 'den 'q --  ) 
 */ 
FCode (p4_mpq_get_den)
{
  mpq_get_den ((mpz_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 2;
}

/**
 * mpq_set_num  ( 'q 'num --  ) 
 */ 
FCode (p4_mpq_set_num)
{
  mpq_set_num ((mpq_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpq_set_den  ( 'q 'den --  ) 
 */ 
FCode (p4_mpq_set_den)
{
  mpq_set_den ((mpq_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**********************
  3.6 Input and output
 **********************/

/**
 * mpq_out_str  ( fid base 'q -- #bytes|0  )
 */
FCode (p4_mpq_out_str)
{
  size_t numbytes = mpq_out_str ( ((p4_File*)SP[2])->f, (int) SP[1],
    (mpq_ptr) SP[0] );
  SP += 2;
  *SP = numbytes ;
}

/**
 * mpq_inp_str  ( 'q fid base|0 -- #bytes|0  )
 */
FCode (p4_mpq_inp_str)
{
  size_t numbytes = mpq_inp_str ((mpq_ptr) SP[2], ((p4_File*)SP[1])->f,
    (int) SP[0]);
  SP += 2;
  *SP = numbytes ;
}

/******************
  4 Floating point
 ******************/

/**
 * /MPF  ( -- sizeof[mpf_t] )
 */

/********************
  4.1 Initialization
 ********************/

/**
 * mpf_set_default_prec  ( prec -- )
 */
FCode (p4_mpf_set_default_prec)
{
  mpf_set_default_prec (*SP++);
}

/**
 * mpf_get_default_prec  ( -- prec )
 */
FCode (p4_mpf_get_default_prec)
{
  *--SP = mpf_get_default_prec ();
}

/**
 * mpf_init  ( 'f -- )
 */
FCode (p4_mpf_init)
{
  mpf_init ((mpf_ptr) *SP++);
}

/**
 * mpf_init2  ( 'f prec -- )
 */
FCode (p4_mpf_init2)
{
  mpf_init2 ((mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/**
 * mpf_clear  ( 'f -- )
 */
FCode (p4_mpf_clear)
{
  mpf_clear ((mpf_ptr) *SP++);
}

/**
 * mpf_get_prec  ( 'f -- prec )
 */
FCode (p4_mpf_get_prec)
{
  *SP = mpf_get_prec ((mpf_ptr) *SP);
}

/**
 * mpf_set_prec  ( 'f prec -- )
 */
FCode (p4_mpf_set_prec)
{
  mpf_set_prec ((mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/**
 * mpf_set_prec_raw  ( 'f prec -- )
 */
FCode (p4_mpf_set_prec_raw)
{
  mpf_set_prec_raw ((mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
}

/****************
  4.2 Assignment
 ****************/

/**
 * mpf_set  ( 'f=f1 'f1 -- )
 */
FCode (p4_mpf_set)
{
  mpf_set ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_set_ui  ( 'f=u u -- )
 */
FCode (p4_mpf_set_ui)
{
  mpf_set_ui ((mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
}

/**
 * mpf_set_si  ( 'f=n n -- )
 */
FCode (p4_mpf_set_si)
{
  mpf_set_si ((mpf_ptr) SP[1], (slong) SP[0]);
  SP += 2;
}

/**
 * mpf_set_d  ( 'f=x f: x -- )
 */
FCode (p4_mpf_set_d)
{
  mpf_set_d ((mpf_ptr) *SP++, *FP++);
}

/**
 * mpf_set_z  ( 'f=b 'b -- )
 */
FCode (p4_mpf_set_z)
{
  mpf_set_z ((mpf_ptr) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_set_q  ( 'f=q 'q -- )
 */
FCode (p4_mpf_set_q)
{
  mpf_set_q ((mpf_ptr) SP[1], (mpq_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_set_str  ( 'f=q 'fzstr base -- )
 */
FCode (p4_mpf_set_str)
{
  int nflag = mpf_set_str ((mpf_ptr) SP[2], (void*) SP[1], (int) SP[0]);
  SP += 2;
  *SP = nflag;
}

/**
 * mpf_swap  ( 'f1 'f2 -- )
 */
FCode (p4_mpf_swap)
{
  mpf_swap ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/********************************************
  4.3 Combined initialization and assignment
 ********************************************/
/**
 * mpf_init_set  ( 'f=f1 'f1 -- )
 */
FCode (p4_mpf_init_set)
{
  mpf_init_set ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_init_set_ui  ( 'f=u u -- )
 */
FCode (p4_mpf_init_set_ui)
{
  mpf_init_set_ui ((mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
}

/**
 * mpf_init_set_si  ( 'f=n n -- )
 */
FCode (p4_mpf_init_set_si)
{
  mpf_init_set_si ((mpf_ptr) SP[1], (slong) SP[0]);
  SP += 2;
}

/**
 * mpf_init_set_d  ( 'f=x f: x -- )
 */
FCode (p4_mpf_init_set_d)
{
  mpf_init_set_d ((mpf_ptr) *SP++, *FP++);
}

/**
 * mpf_init_set_str  ( 'f=q 'fzstr base -- )
 */
FCode (p4_mpf_init_set_str)
{
  int nflag = mpf_init_set_str ((mpf_ptr) SP[2], (void*) SP[1], (int) SP[0]);
  SP += 2;
  *SP = nflag;
}

/****************
  4.4 Conversion
 ****************/

/**
 * mpf_get_d  ( 'f -- f: f )
 */
FCode (p4_mpf_get_d)
{
  *--FP = mpf_get_d ((mpf_ptr) *SP++);
}

/**
 * mpf_get_d_2exp  ( 'exp 'f -- f: frac )
 */
FCode (p4_mpf_get_d_2exp)
{
  *--FP = mpf_get_d_2exp ((signed long int*) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_get_si  ( 'f -- [f] )
 */
FCode (p4_mpf_get_si)
{
  *SP = (p4cell) mpf_get_si ((mpf_ptr) *SP);
}

/**
 * mpf_get_ui  ( 'f -- [f] )
 */
FCode (p4_mpf_get_ui)
{
  *SP = (p4ucell) mpf_get_ui ((mpf_ptr) *SP);
}

/**
 * mpf_get_str  ( 'fzstr|0 'exp base #digits 'f -- 'fzstr|'fzstr' )
 */
FCode (p4_mpf_get_str)
{
  char* fzstr = mpf_get_str ( (char*) SP[4], (mp_exp_t*) SP[3],
    (int) SP[2], (size_t) SP[1], (mpf_ptr) SP[0]);
  SP += 4;
  *SP = (p4ucell) fzstr;
}

/****************
  4.5 Arithmetic
 ****************/

/**
 * mpf_add  ( 'f1+f2 'f1 'f2 -- )
 */
FCode (p4_mpf_add)
{
  mpf_add ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_add_ui  ( 'f+u 'f u -- )
 */
FCode (p4_mpf_add_ui)
{
  mpf_add_ui ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpf_sub  ( 'f1-f2 'f1 'f2 -- )
 */
FCode (p4_mpf_sub)
{
  mpf_sub ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_ui-sub  ( 'u-f u 'f -- )
 */
FCode (p4_mpf_ui_sub)
{
  mpf_ui_sub ((mpf_ptr) SP[2], (ulong) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_sub_ui  ( 'f-u 'f u -- )
 */
FCode (p4_mpf_sub_ui)
{
  mpf_sub_ui ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpf_mul  ( 'f1*f2 'f1 'f2 -- )
 */
FCode (p4_mpf_mul)
{
  mpf_mul ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_mul_ui  ( 'f*u 'f u -- )
 */
FCode (p4_mpf_mul_ui)
{
  mpf_mul_ui ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpf_div  ( 'f1/f2 'f1 'f2 -- )
 */
FCode (p4_mpf_div)
{
  mpf_div ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_ui_div  ( 'u/f u 'f -- )
 */
FCode (p4_mpf_ui_div)
{
  mpf_ui_div ((mpf_ptr) SP[2], (ulong) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_div_ui  ( 'f/u 'f u -- )
 */
FCode (p4_mpf_div_ui)
{
  mpf_div_ui ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpf_sqrt  ( 'root 'f -- )
 */
FCode (p4_mpf_sqrt)
{
  mpf_sqrt ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_sqrt_ui  ( 'root u -- )
 */
FCode (p4_mpf_sqrt_ui)
{
  mpf_sqrt_ui ((mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 2;
}

/**
 * mpf_pow_ui  ( 'root 'f u -- )
 */
FCode (p4_mpf_pow_ui)
{
  mpf_pow_ui ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (ulong) SP[0]);
  SP += 3;
}

/**
 * mpf_neg  ( '-f 'f -- )
 */
FCode (p4_mpf_neg)
{
  mpf_neg ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_abs  ( '|f| 'f -- )
 */
FCode (p4_mpf_abs)
{
  mpf_abs ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_mul_2exp  ( 'f*2^p 'f +p -- )
 */
FCode (p4_mpf_mul_2exp)
{
  mpf_mul_2exp ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpf_div_2exp  ( 'f/2^p 'f +p -- )
 */
FCode (p4_mpf_div_2exp)
{
  mpf_div_2exp ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/****************
  4.6 Comparison
 ****************/

/**
 * mpf_cmp  ( 'f1 'f2 -- +n|0|-n )
 * +n: f1 > f2
 *  0: f1 = f2
 * -n: f1 < f2
 */
FCode (p4_mpf_cmp)
{
  int n = mpf_cmp ( (mpf_ptr) SP[1], (mpf_ptr) SP[0] );
  *++SP = n;
}

/**
 * mpf_cmp_d  ( 'f f: x -- s: +n|0|-n )
 * +n: f1 > f2
 *  0: f1 = f2
 * -n: f1 < f2
 */
FCode (p4_mpf_cmp_d)
{
  int n = mpf_cmp_d ( (mpf_ptr) *SP, *FP++ );
  *SP = n;
}

/**
 * mpf_cmp_ui  ( 'f u -- +n'|0|-n' )
 * +n: f > u
 *  0: f = u
 * -n: f < u
 */
FCode (p4_mpf_cmp_ui)
{
  int n = mpf_cmp_ui ( (mpf_ptr) SP[1], (ulong) SP[0]);
  *++SP = n;
}

/**
 * mpf_cmp_si  ( 'f n -- +n'|0|-n' )
 * +n: f > n
 *  0: f = n
 * -n: f < n
 */
FCode (p4_mpf_cmp_si)
{
  int n = mpf_cmp_si ( (mpf_ptr) SP[1], (slong) SP[0]);
  *++SP = n;
}

/**
 * mpf_eq  ( 'f1 'f2 #bits -- flag )
 */
FCode (p4_mpf_eq)
{
  int flag = mpf_eq ( (mpf_ptr) SP[2], (mpf_ptr) SP[1], (mp_bitcnt_t) SP[0]);
  SP += 2;
  *SP = flag ? ~0 : 0;
}

/**
 * mpf_reldiff  ( '|f1-f2|/f1 'f1 'f2 -- )
 */
FCode (p4_mpf_reldiff)
{
  mpf_reldiff ((mpf_ptr) SP[2], (mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 3;
}

/**
 * mpf_sgn  ( 'f -- flag )
 */
FCode (p4_mpf_sgn)
{
  *SP = mpf_sgn ((mpz_ptr) *SP);
}

/**********************
  4.7 Input and output
 **********************/

/**
 * mpf_out_str  ( fid base #digits 'f -- #bytes|0  )
 */
FCode (p4_mpf_out_str)
{
  size_t numbytes = mpf_out_str ( ((p4_File*)SP[3])->f, (int) SP[2],
    (size_t) SP[1], (mpf_ptr) SP[0] );
  SP += 3;
  *SP = numbytes ;
}

/**
 * mpf_inp_str  ( 'f fid base|0 -- #bytes|0  )
 */
FCode (p4_mpf_inp_str)
{
  size_t numbytes = mpf_inp_str ((mpf_ptr) SP[2], ((p4_File*)SP[1])->f,
    (int) SP[0]);
  SP += 2;
  *SP = numbytes ;
}

/*******************
  4.8 Miscellaneous
 *******************/

/**
 * mpf_ceil  ( 'c[f] 'f -- )
 */
FCode (p4_mpf_ceil)
{
  mpf_ceil ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_floor  ( 'f[f] 'f -- )
 */
FCode (p4_mpf_floor)
{
  mpf_floor ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_trunc  ( 't[f] 'f -- )
 */
FCode (p4_mpf_trunc)
{
  mpf_trunc ((mpf_ptr) SP[1], (mpf_ptr) SP[0]);
  SP += 2;
}

/**
 * mpf_ingteger_p  ( 'f -- flag )
 */
FCode (p4_mpf_integer_p)
{
  *SP = mpf_integer_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_ulong_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_ulong_p)
{
  *SP = mpf_fits_ulong_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_slong_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_slong_p)
{
  *SP = mpf_fits_slong_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_uint_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_uint_p)
{
  *SP = mpf_fits_uint_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_sint_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_sint_p)
{
  *SP = mpf_fits_sint_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_ushort_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_ushort_p)
{
  *SP = mpf_fits_ushort_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_fits_sshort_p  ( 'f -- flag )
 */
FCode (p4_mpf_fits_sshort_p)
{
  *SP = mpf_fits_sshort_p ((mpf_ptr) *SP) ? ~0 : 0;
}

/**
 * mpf_urandomb  ( 'f 'ranstate #bits -- )
 */
FCode (p4_mpf_urandomb)
{
  mpf_urandomb ((mpf_ptr) SP[2], (gmp_rstate_t) SP[1],
    (mp_bitcnt_t) SP[0]);
  SP += 3;
}

/**
 * mpf_random2  ( 'f maxsize exp -- )
 */
FCode (p4_mpf_random2)
{
  mpf_random2 ((mpf_ptr) SP[2], (mp_size_t) SP[1], (mp_exp_t) SP[0]);
  SP += 3;
}

/************
 5 Low level
 ************/

/***********************
  6 Random number state
 ***********************/

/**
 * /gmp_randstate  ( -- sizeof[gmp_randstate_t] )
 */

/********************
  6.1 Initialization
 ********************/

/**
 * gmp_randinit_default  ( 'state -- )
 */
FCode (p4_gmp_randinit_default)
{
  gmp_randinit_default ((gmp_rstate_t) *SP++);
}

/**
 * gmp_randinit_mt  ( 'state -- )
 */
FCode (p4_gmp_randinit_mt)
{
  gmp_randinit_mt ((gmp_rstate_t) *SP++);
}

/**
 * gmp_randinit_lc_2exp  ( 'state 'a +c +m2exp -- )
 */
FCode (p4_gmp_randinit_lc_2exp)
{
  gmp_randinit_lc_2exp ((gmp_rstate_t) SP[3], (mpz_ptr) SP[2],
    (ulong) SP[1], (ulong) SP[0]);
  SP += 4;
}

/**
 * gmp_randinit_lc_2exp_siZE  ( 'state #bits -- flag  )
 */
FCode (p4_gmp_randinit_lc_2exp_size)
{
  ulong flag = gmp_randinit_lc_2exp_size ((gmp_rstate_t) SP[1],
    (ulong) SP[0]);
  *++SP = flag ? ~0 : 0 ;
}

/**
 * gmp_randinit_set  ( 'state.copy 'state -- )
 */
FCode (p4_gmp_randinit_set)
{
  gmp_randinit_set ((gmp_rstate_t) SP[1], (gmp_rstate_t) SP[0]);
  SP += 2;
}

/**
 * gmp_randclear  ( 'state -- )
 */
FCode (p4_gmp_randclear)
{
  gmp_randclear ((gmp_rstate_t) *SP++);
}

/*************
  6.2 Seeding
 *************/

/**
 * gmp_randseed  ( 'state.copy 'seed -- )
 */
FCode (p4_gmp_randseed)
{
  gmp_randseed ((gmp_rstate_t) SP[1], (mpz_ptr) SP[0]);
  SP += 2;
}

/**
 * gmp_randseed_ui  ( 'state.copy +seed -- )
 */
FCode (p4_gmp_randseed_ui)
{
  gmp_randseed_ui ((gmp_rstate_t) SP[1], (ulong) SP[0]);
  SP += 2;
}

/*******************
  6.3 Miscellaneous
 *******************/

/**
 * gmp_urandomb_ui  ( 'state #bits -- ran )  \ #bits <= 32
 */
FCode (p4_gmp_urandomb_ui)
{
  ulong ran = gmp_urandomb_ui ((gmp_rstate_t) SP[1], (ulong) SP[0]);
  *++SP = ran;
}

/**
 * gmp_urandomm_ui  ( 'state #bits -- ran )  \ #bits <= 32
 */
FCode (p4_gmp_urandomm_ui)
{
  ulong ran = gmp_urandomm_ui ((gmp_rstate_t) SP[1], (ulong) SP[0]);
  *++SP = ran;
}


/*******************
 7 Formatted output
 *******************/
/******************
 8 Formatted input
 ******************/


/*******
  Words
 *******/

P4_LISTWORDS (libgmp_pfe) =
{
// 1 General support
   P4_NEED ("floating-ext"),
   P4_FXco ("gmp_version",	p4_gmp_version),
   P4_FXco ("mp_bits_per_limb",	p4_mp_bits_per_limb),
// 2 Integers
   P4_OCoN ("/MPZ",		sizeof (mpz_t)),
// 2.1 Initialization
   P4_FXco ("mpz_init",		p4_mpz_init),
   P4_FXco ("mpz_clear",	p4_mpz_clear),
// 2.2 Assignment
   P4_FXco ("mpz_set",		p4_mpz_set),
   P4_FXco ("mpz_set_ui",	p4_mpz_set_ui),
   P4_FXco ("mpz_set_si",	p4_mpz_set_si),
   P4_FXco ("mpz_set_d",	p4_mpz_set_d),
   P4_FXco ("mpz_set_q",	p4_mpz_set_q),
   P4_FXco ("mpz_set_f",	p4_mpz_set_f),
   P4_FXco ("mpz_set_str",	p4_mpz_set_str),
   P4_FXco ("mpz_swap",		p4_mpz_swap),
// 2.3 Combined initialization and assignment
   P4_FXco ("mpz_init_set",	p4_mpz_init_set),
   P4_FXco ("mpz_init_set_ui",	p4_mpz_init_set_ui),
   P4_FXco ("mpz_init_set_si",	p4_mpz_init_set_si),
   P4_FXco ("mpz_init_set_d",	p4_mpz_init_set_d),
   P4_FXco ("mpz_init_set_str",	p4_mpz_init_set_str),
// 2.4 Conversion
   P4_FXco ("mpz_get_ui",	p4_mpz_get_ui),
   P4_FXco ("mpz_get_si",	p4_mpz_get_si),
   P4_FXco ("mpz_get_d",	p4_mpz_get_d),
   P4_FXco ("mpz_get_d_2exp",	p4_mpz_get_d_2exp),
   P4_FXco ("mpz_get_str",	p4_mpz_get_str),
// 2.5 Arithmetic
   P4_FXco ("mpz_add",		p4_mpz_add),
   P4_FXco ("mpz_add_ui",	p4_mpz_add_ui),
   P4_FXco ("mpz_sub",		p4_mpz_sub),
   P4_FXco ("mpz_sub_ui",	p4_mpz_sub_ui),
   P4_FXco ("mpz_ui_sub",	p4_mpz_ui_sub),
   P4_FXco ("mpz_mul",		p4_mpz_mul),
   P4_FXco ("mpz_mul_si",	p4_mpz_mul_si),
   P4_FXco ("mpz_mul_ui",	p4_mpz_mul_ui),
   P4_FXco ("mpz_addmul",	p4_mpz_addmul),
   P4_FXco ("mpz_addmul_ui",	p4_mpz_addmul_ui),
   P4_FXco ("mpz_submul",	p4_mpz_submul),
   P4_FXco ("mpz_submul_ui",	p4_mpz_submul_ui),
   P4_FXco ("mpz_mul_2exp",	p4_mpz_mul_2exp),
   P4_FXco ("mpz_neg",		p4_mpz_neg),
   P4_FXco ("mpz_abs",		p4_mpz_abs),
// 2.6 Division
   P4_FXco ("mpz_cdiv_q",	p4_mpz_cdiv_q),
   P4_FXco ("mpz_cdiv_r",	p4_mpz_cdiv_r),
   P4_FXco ("mpz_cdiv_qr",	p4_mpz_cdiv_qr),
   P4_FXco ("mpz_cdiv_q_ui",	p4_mpz_cdiv_q_ui),
   P4_FXco ("mpz_cdiv_r_ui",	p4_mpz_cdiv_r_ui),
   P4_FXco ("mpz_cdiv_qr_ui",	p4_mpz_cdiv_qr_ui),
   P4_FXco ("mpz_cdiv_ui",	p4_mpz_cdiv_ui),
   P4_FXco ("mpz_cdiv_q_2exp",	p4_mpz_cdiv_q_2exp),
   P4_FXco ("mpz_cdiv_r_2exp",	p4_mpz_cdiv_r_2exp),
   P4_FXco ("mpz_fdiv_q",	p4_mpz_fdiv_q),
   P4_FXco ("mpz_fdiv_r",	p4_mpz_fdiv_r),
   P4_FXco ("mpz_fdiv_qr",	p4_mpz_fdiv_qr),
   P4_FXco ("mpz_fdiv_q_ui",	p4_mpz_fdiv_q_ui),
   P4_FXco ("mpz_fdiv_r_ui",	p4_mpz_fdiv_r_ui),
   P4_FXco ("mpz_fdiv_qr_ui",	p4_mpz_fdiv_qr_ui),
   P4_FXco ("mpz_fdiv_ui",	p4_mpz_fdiv_ui),
   P4_FXco ("mpz_fdiv_q_2exp",	p4_mpz_fdiv_q_2exp),
   P4_FXco ("mpz_fdiv_r_2exp",	p4_mpz_fdiv_r_2exp),
   P4_FXco ("mpz_tdiv_q",	p4_mpz_tdiv_q),
   P4_FXco ("mpz_tdiv_r",	p4_mpz_tdiv_r),
   P4_FXco ("mpz_tdiv_qr",	p4_mpz_tdiv_qr),
   P4_FXco ("mpz_tdiv_q_ui",	p4_mpz_tdiv_q_ui),
   P4_FXco ("mpz_tdiv_r_ui",	p4_mpz_tdiv_r_ui),
   P4_FXco ("mpz_tdiv_qr_ui",	p4_mpz_tdiv_qr_ui),
   P4_FXco ("mpz_tdiv_ui",	p4_mpz_tdiv_ui),
   P4_FXco ("mpz_tdiv_q_2exp",	p4_mpz_tdiv_q_2exp),
   P4_FXco ("mpz_tdiv_r_2exp",	p4_mpz_tdiv_r_2exp),
   P4_FXco ("mpz_mod",		p4_mpz_mod),
   P4_FXco ("mpz_mod_ui",	p4_mpz_mod_ui),
   P4_FXco ("mpz_divexact",	p4_mpz_divexact),
   P4_FXco ("mpz_divexact_ui",	p4_mpz_divexact_ui),
   P4_FXco ("mpz_divisible_p",	p4_mpz_divisible_p),
   P4_FXco ("mpz_divisible_ui_p",    p4_mpz_divisible_ui_p),
   P4_FXco ("mpz_divisible_2exp_p",  p4_mpz_divisible_2exp_p),
   P4_FXco ("mpz_congruent_p",	p4_mpz_congruent_p),
   P4_FXco ("mpz_congruent_ui_p",    p4_mpz_congruent_ui_p),
   P4_FXco ("mpz_congruent_2exp_p",  p4_mpz_congruent_2exp_p),
// 2.7  Exponentiation
   P4_FXco ("mpz_powm",		p4_mpz_powm),
   P4_FXco ("mpz_powm_ui",	p4_mpz_powm_ui),
   P4_FXco ("mpz_powm_sec",	p4_mpz_powm_sec),
   P4_FXco ("mpz_pow_ui",	p4_mpz_pow_ui),
   P4_FXco ("mpz_ui_pow_ui",	p4_mpz_ui_pow_ui),
// 2.8  Root extraction
   P4_FXco ("mpz_root",		p4_mpz_root),
   P4_FXco ("mpz_rootrem",	p4_mpz_rootrem),
   P4_FXco ("mpz_sqrt",		p4_mpz_sqrt),
   P4_FXco ("mpz_sqrtrem",	p4_mpz_sqrtrem),
   P4_FXco ("mpz_perfect_power_p",	p4_mpz_perfect_power_p),
   P4_FXco ("mpz_perfect_square_p",	p4_mpz_perfect_square_p),
// 2.9  Number theoretics
   P4_FXco ("mpz_probab_prime_p",	p4_mpz_probab_prime_p),
   P4_FXco ("mpz_nextprime",	p4_mpz_nextprime),
   P4_FXco ("mpz_gcd",		p4_mpz_gcd),
   P4_FXco ("mpz_gcd_ui",	p4_mpz_gcd_ui),
   P4_FXco ("mpz_gcdext",	p4_mpz_gcdext),
   P4_FXco ("mpz_lcm",		p4_mpz_lcm),
   P4_FXco ("mpz_lcm_ui",	p4_mpz_lcm_ui),
   P4_FXco ("mpz_invert",	p4_mpz_invert),
   P4_FXco ("mpz_jacobi",	p4_mpz_jacobi),
   P4_FXco ("mpz_legendre",	p4_mpz_legendre),
   P4_FXco ("mpz_kronecker",	p4_mpz_kronecker),
   P4_FXco ("mpz_kronecker_si",	p4_mpz_kronecker_si),
   P4_FXco ("mpz_kronecker_ui",	p4_mpz_kronecker_ui),
   P4_FXco ("mpz_si_kronecker",	p4_mpz_si_kronecker),
   P4_FXco ("mpz_ui_kronecker",	p4_mpz_ui_kronecker),
   P4_FXco ("mpz_remove",	p4_mpz_remove),
   P4_FXco ("mpz_fac_ui",	p4_mpz_fac_ui),
   P4_FXco ("mpz_bin_ui",	p4_mpz_bin_ui),
   P4_FXco ("mpz_bin_uiui",	p4_mpz_bin_uiui),
   P4_FXco ("mpz_fib_ui",	p4_mpz_fib_ui),
   P4_FXco ("mpz_fib2_ui",	p4_mpz_fib2_ui),
   P4_FXco ("mpz_lucnum_ui",	p4_mpz_lucnum_ui),
   P4_FXco ("mpz_lucnum2_ui",	p4_mpz_lucnum2_ui),
// 2.10 Comparison
   P4_FXco ("mpz_cmp",		p4_mpz_cmp),
   P4_FXco ("mpz_cmp_d",	p4_mpz_cmp_d),
   P4_FXco ("mpz_cmp_si",	p4_mpz_cmp_si),
   P4_FXco ("mpz_cmp_ui",	p4_mpz_cmp_ui),
   P4_FXco ("mpz_cmpabs",	p4_mpz_cmpabs),
   P4_FXco ("mpz_cmpabs_d",	p4_mpz_cmpabs_d),
   P4_FXco ("mpz_cmpabs_ui",	p4_mpz_cmpabs_ui),
   P4_FXco ("mpz_sgn",		p4_mpz_sgn),
// 2.11 Logic and bit manipulation
   P4_FXco ("mpz_and",		p4_mpz_and),
   P4_FXco ("mpz_ior",		p4_mpz_ior),
   P4_FXco ("mpz_xor",		p4_mpz_xor),
   P4_FXco ("mpz_com",		p4_mpz_com),
   P4_FXco ("mpz_popcount",	p4_mpz_popcount),
   P4_FXco ("mpz_hamdist",	p4_mpz_hamdist),
   P4_FXco ("mpz_scan0",	p4_mpz_scan0),
   P4_FXco ("mpz_scan1",	p4_mpz_scan1),
   P4_FXco ("mpz_setbit",	p4_mpz_setbit),
   P4_FXco ("mpz_clrbit",	p4_mpz_clrbit),
   P4_FXco ("mpz_combit",	p4_mpz_combit),
   P4_FXco ("mpz_tstbit",	p4_mpz_tstbit),
// 2.12 Input and output
   P4_FXco ("mpz_out_str",	p4_mpz_out_str),
   P4_FXco ("mpz_inp_str",	p4_mpz_inp_str),
   P4_FXco ("mpz_out_raw",	p4_mpz_out_raw),
   P4_FXco ("mpz_inp_raw",	p4_mpz_inp_raw),
// 2.13 Random numbers
   P4_FXco ("mpz_urandomb",	p4_mpz_urandomb),
   P4_FXco ("mpz_urandomm",	p4_mpz_urandomm),
   P4_FXco ("mpz_rrandomb",	p4_mpz_rrandomb),
// 2.14 Integer export and import
   P4_FXco ("mpz_import",	p4_mpz_import),
   P4_FXco ("mpz_export",	p4_mpz_export),
// 2.15 Miscellaneous
   P4_FXco ("mpz_fits_ulong_p",	p4_mpz_fits_ulong_p),
   P4_FXco ("mpz_fits_slong_p",	p4_mpz_fits_slong_p),
   P4_FXco ("mpz_fits_uint_p",	p4_mpz_fits_uint_p),
   P4_FXco ("mpz_fits_sint_p",	p4_mpz_fits_sint_p),
   P4_FXco ("mpz_fits_ushort_p",	p4_mpz_fits_ushort_p),
   P4_FXco ("mpz_fits_sshort_p",	p4_mpz_fits_sshort_p),
   P4_FXco ("mpz_odd_p",	p4_mpz_odd_p),
   P4_FXco ("mpz_even_p",	p4_mpz_even_p),
   P4_FXco ("mpz_sizeinbase",	p4_mpz_sizeinbase),
// 2.16 Special
   P4_FXco ("mpz_getlimbn",	p4_mpz_getlimbn),
   P4_FXco ("mpz_size",		p4_mpz_size),
// 3 Rationals
   P4_OCoN ("/MPQ",		sizeof (mpq_t)),
   P4_FXco ("mpq_canonicalize",	p4_mpq_canonicalize),
// 3.1 Initialization and assignment
   P4_FXco ("mpq_init",		p4_mpq_init),
   P4_FXco ("mpq_clear",	p4_mpq_clear),
   P4_FXco ("mpq_set",		p4_mpq_set),
   P4_FXco ("mpq_set_z",	p4_mpq_set_z),
   P4_FXco ("mpq_set_ui",	p4_mpq_set_ui),
   P4_FXco ("mpq_set_si",	p4_mpq_set_si),
   P4_FXco ("mpq_set_str",	p4_mpq_set_str),
   P4_FXco ("mpq_swap",		p4_mpq_swap),
// 3.2 Conversion
   P4_FXco ("mpq_get_d",	p4_mpq_get_d),
   P4_FXco ("mpq_set_d",	p4_mpq_set_d),
   P4_FXco ("mpq_set_f",	p4_mpq_set_f),
   P4_FXco ("mpq_get_str",	p4_mpq_get_str),
// 3.3 Arithmetic
   P4_FXco ("mpq_add",		p4_mpq_add),
   P4_FXco ("mpq_sub",		p4_mpq_sub),
   P4_FXco ("mpq_mul",		p4_mpq_mul),
   P4_FXco ("mpq_mul_2exp",	p4_mpq_mul_2exp),
   P4_FXco ("mpq_div",		p4_mpq_div),
   P4_FXco ("mpq_div_2exp",	p4_mpq_div_2exp),
   P4_FXco ("mpq_neg",		p4_mpq_neg),
   P4_FXco ("mpq_abs",		p4_mpq_abs),
   P4_FXco ("mpq_inv",		p4_mpq_inv),
// 3.4 Comparison
   P4_FXco ("mpq_cmp",		p4_mpq_cmp),
   P4_FXco ("mpq_cmp_ui",	p4_mpq_cmp_ui),
   P4_FXco ("mpq_cmp_si",	p4_mpq_cmp_si),
   P4_FXco ("mpq_sgn",		p4_mpq_sgn),
   P4_FXco ("mpq_equal",	p4_mpq_equal),
// 3.5 Integer functions
   P4_FXco ("mpq_numref",	p4_mpq_numref),
   P4_FXco ("mpq_denref",	p4_mpq_denref),
   P4_FXco ("mpq_get_num",	p4_mpq_get_num),
   P4_FXco ("mpq_get_den",	p4_mpq_get_den),
   P4_FXco ("mpq_set_num",	p4_mpq_set_num),
   P4_FXco ("mpq_set_den",	p4_mpq_set_den),
// 3.6 Input and output
   P4_FXco ("mpq_out_str",	p4_mpq_out_str),
   P4_FXco ("mpq_inp_str",	p4_mpq_inp_str),
// 4 Floating point
   P4_OCoN ("/MPF",		sizeof (mpf_t)),
// 4.1 Initialization
   P4_FXco ("mpf_set_default_prec",	p4_mpf_set_default_prec),
   P4_FXco ("mpf_get_default_prec",	p4_mpf_get_default_prec),
   P4_FXco ("mpf_init",		p4_mpf_init),
   P4_FXco ("mpf_init2",	p4_mpf_init2),
   P4_FXco ("mpf_clear",	p4_mpf_clear),
   P4_FXco ("mpf_get_prec",	p4_mpf_get_prec),
   P4_FXco ("mpf_set_prec",	p4_mpf_set_prec),
   P4_FXco ("mpf_set_prec_raw",	p4_mpf_set_prec_raw),
// 4.2 Assignment
   P4_FXco ("mpf_set",		p4_mpf_set),
   P4_FXco ("mpf_set_ui",	p4_mpf_set_ui),
   P4_FXco ("mpf_set_si",	p4_mpf_set_si),
   P4_FXco ("mpf_set_d",	p4_mpf_set_d),
   P4_FXco ("mpf_set_z",	p4_mpf_set_z),
   P4_FXco ("mpf_set_q",	p4_mpf_set_q),
   P4_FXco ("mpf_set_str",	p4_mpf_set_str),
   P4_FXco ("mpf_swap",		p4_mpf_swap),
// 4.3 Combined initialization and assignment
   P4_FXco ("mpf_init_set",	p4_mpf_init_set),
   P4_FXco ("mpf_init_set_ui",	p4_mpf_init_set_ui),
   P4_FXco ("mpf_init_set_si",	p4_mpf_init_set_si),
   P4_FXco ("mpf_init_set_d",	p4_mpf_init_set_d),
   P4_FXco ("mpf_init_set_str",	p4_mpf_init_set_str),
// 4.4 Conversion
// P4_OCoN ("/mp_exp_t",	sizeof(mp_exp_t)), // debugging
   P4_FXco ("mpf_get_d",	p4_mpf_get_d),
   P4_FXco ("mpf_get_d_2exp",	p4_mpf_get_d_2exp),
   P4_FXco ("mpf_get_si",	p4_mpf_get_si),
   P4_FXco ("mpf_get_ui",	p4_mpf_get_ui),
   P4_FXco ("mpf_get_str",	p4_mpf_get_str),
// 4.5 Arithmetic
   P4_FXco ("mpf_add",		p4_mpf_add),
   P4_FXco ("mpf_add_ui",	p4_mpf_add_ui),
   P4_FXco ("mpf_sub",		p4_mpf_sub),
   P4_FXco ("mpf_ui_sub",	p4_mpf_ui_sub),
   P4_FXco ("mpf_sub_ui",	p4_mpf_sub_ui),
   P4_FXco ("mpf_mul",		p4_mpf_mul),
   P4_FXco ("mpf_mul_ui",	p4_mpf_mul_ui),
   P4_FXco ("mpf_div",		p4_mpf_div),
   P4_FXco ("mpf_ui_div",	p4_mpf_ui_div),
   P4_FXco ("mpf_div_ui",	p4_mpf_div_ui),
   P4_FXco ("mpf_sqrt",		p4_mpf_sqrt),
   P4_FXco ("mpf_sqrt_ui",	p4_mpf_sqrt_ui),
   P4_FXco ("mpf_pow_ui",	p4_mpf_pow_ui),
   P4_FXco ("mpf_neg",		p4_mpf_neg),
   P4_FXco ("mpf_abs",		p4_mpf_abs),
   P4_FXco ("mpf_mul_2exp",	p4_mpf_mul_2exp),
   P4_FXco ("mpf_div_2exp",	p4_mpf_div_2exp),
// 4.6 Comparison
   P4_FXco ("mpf_cmp",		p4_mpf_cmp),
   P4_FXco ("mpf_cmp_d",	p4_mpf_cmp_d),
   P4_FXco ("mpf_cmp_ui",	p4_mpf_cmp_ui),
   P4_FXco ("mpf_cmp_si",	p4_mpf_cmp_si),
   P4_FXco ("mpf_eq",		p4_mpf_eq),
   P4_FXco ("mpf_reldiff",	p4_mpf_reldiff),
   P4_FXco ("mpf_sgn",		p4_mpf_sgn),
// 4.7 Input and output
   P4_FXco ("mpf_out_str",	p4_mpf_out_str),
   P4_FXco ("mpf_inp_str",	p4_mpf_inp_str),
// 4.8 Miscellaneous
   P4_FXco ("mpf_ceil",		p4_mpf_ceil),
   P4_FXco ("mpf_floor",	p4_mpf_floor),
   P4_FXco ("mpf_trunc",	p4_mpf_trunc),
   P4_FXco ("mpf_integer_p",	p4_mpf_integer_p),
   P4_FXco ("mpf_fits_ulong_p",	p4_mpf_fits_ulong_p),
   P4_FXco ("mpf_fits_slong_p",	p4_mpf_fits_slong_p),
   P4_FXco ("mpf_fits_uint_p",	p4_mpf_fits_uint_p),
   P4_FXco ("mpf_fits_sint_p",	p4_mpf_fits_sint_p),
   P4_FXco ("mpf_fits_ushort_p",	p4_mpf_fits_ushort_p),
   P4_FXco ("mpf_fits_sshort_p",	p4_mpf_fits_sshort_p),
   P4_FXco ("mpf_urandomb",	p4_mpf_urandomb),
   P4_FXco ("mpf_random2",	p4_mpf_random2),
// 5 Low level
// 6 Random number state
   P4_OCoN ("/GMP-RANDSTATE",	sizeof (gmp_randstate_t)),
// 6.1 Initialization
   P4_FXco ("gmp_randinit_default",    p4_gmp_randinit_default),
   P4_FXco ("gmp_randinit_mt",	p4_gmp_randinit_mt),
   P4_FXco ("gmp_randinit_lc_2exp",	p4_gmp_randinit_lc_2exp),
   P4_FXco ("gmp_randinit_lc_2exp_size",    p4_gmp_randinit_lc_2exp_size),
   P4_FXco ("gmp_randinit_set",	p4_gmp_randinit_set),
   P4_FXco ("gmp_randclear",	p4_gmp_randclear),
// 6.2 Seeding
   P4_FXco ("gmp_randseed",	p4_gmp_randseed),
   P4_FXco ("gmp_randseed_ui",	p4_gmp_randseed_ui),
// 6.3 Miscellaneous
   P4_FXco ("gmp_urandomb_ui",	p4_gmp_urandomb_ui),
   P4_FXco ("gmp_urandomm_ui",	p4_gmp_urandomm_ui),
// 7 Formatted output
// 8 Formatted input
};

P4_COUNTWORDS (libgmp_pfe, "GMP bindings for pfe");
